# 实现复杂联动逻辑

我们的联动逻辑的复杂度通常取决于是否存在以下几种问题：

- 联动类型复杂
  - 一对多联动
  - 多对一联动
  - 多依赖联动
  - 链式联动
  - 循环联动
  - 联动校验
- 联动过程中存在异步
- 联动过程中存在大量数据转换逻辑
- 自增列表中相邻字段或者与上一级/下一级字段间的联动
- 动态联动，这里主要讲的是 JSON Schema 中的动态联动，比如是用户手工配置联动规则

上面分析那么多，其实每种问题都是比较麻烦的，一不小心，就会把代码变得非常难以维护，所以我们在解决这类复杂问题的时候，更多的是需要寻找一个可以优雅分解问题的解题策略。

直接看例子：

## 一对多联动

```jsx
import React, { useEffect } from 'react'
import ReactDOM from 'react-dom'
import {
  SchemaForm,
  SchemaMarkupField as Field,
  FormButtonGroup,
  createFormActions,
  FormEffectHooks,
  Submit,
  Reset
} from '@formily/antd' // 或者 @formily/next
import { Input, Select } from '@formily/antd-components'
import Printer from '@formily/printer'
import 'antd/dist/antd.css'

const { onFieldValueChange$ } = FormEffectHooks

const useOneToManyEffects = () => {
  const { setFieldState } = createFormActions()
  onFieldValueChange$('aa').subscribe(({ value }) => {
    setFieldState('*(bb,cc,dd)', state => {
      state.visible = value
    })
  })
}

const App = () => {
  return (
    <Printer>
      <SchemaForm
        components={{ Input, Select }}
        onSubmit={values => {
          console.log(values)
        }}
        effects={() => {
          useOneToManyEffects()
        }}
      >
        <Field
          type="string"
          enum={[
            { label: 'visible', value: true },
            { label: 'hidden', value: false }
          ]}
          default={false}
          name="aa"
          title="AA"
          x-component="Select"
        />
        <Field type="string" name="bb" title="BB" x-component="Input" />
        <Field type="string" name="cc" title="CC" x-component="Input" />
        <Field type="string" name="dd" title="DD" x-component="Input" />
      </SchemaForm>
    </Printer>
  )
}

ReactDOM.render(<App />, document.getElementById('root'))
```

**案例解析**

- 使用 FormEffectHooks 可以很方便的将联动逻辑拆分出去，方便我们进行物理分离
- 借助路径系统的批量匹配能力实现一对多联动

## 自己联动自己(副作用校验)

```jsx
import React, { useEffect } from 'react'
import ReactDOM from 'react-dom'
import {
  SchemaForm,
  SchemaMarkupField as Field,
  FormButtonGroup,
  createFormActions,
  FormEffectHooks,
  Submit,
  Reset
} from '@formily/antd' // 或者 @formily/next
import { Input, Select } from '@formily/antd-components'
import Printer from '@formily/printer'
import { merge } from 'rxjs'
import 'antd/dist/antd.css'

const { onFieldValueChange$, onFormSubmitStart$ } = FormEffectHooks

const useValidator = (name, validator) => {
  const { setFieldState } = createFormActions()
  merge(onFieldValueChange$(name), onFormSubmitStart$()).subscribe(() => {
    setFieldState(name, state => {
      state.errors = validator(state.value, state)
    })
  })
}

const App = () => {
  return (
    <Printer>
      <SchemaForm
        components={{ Input, Select }}
        onSubmit={values => {
          console.log(values)
        }}
        effects={() => {
          useValidator('aa', value => (!value ? '必填字段' : ''))
        }}
      >
        <Field type="string" name="aa" title="AA" x-component="Input" />
        <FormButtonGroup>
          <Submit />
        </FormButtonGroup>
      </SchemaForm>
    </Printer>
  )
}

ReactDOM.render(<App />, document.getElementById('root'))
```

## 多对一联动

```jsx
import React, { useEffect } from 'react'
import ReactDOM from 'react-dom'
import {
  SchemaForm,
  SchemaMarkupField as Field,
  FormButtonGroup,
  createFormActions,
  FormEffectHooks,
  Submit,
  Reset
} from '@formily/antd' // 或者 @formily/next
import { Input, Select } from '@formily/antd-components'
import Printer from '@formily/printer'
import 'antd/dist/antd.css'

const { onFieldValueChange$ } = FormEffectHooks

const useManyToOneEffects = () => {
  const { setFieldState } = createFormActions()
  onFieldValueChange$('bb').subscribe(({ value }) => {
    setFieldState('aa', state => {
      state.visible = value
    })
  })
  onFieldValueChange$('cc').subscribe(({ value }) => {
    setFieldState('aa', state => {
      state.value = value
    })
  })
}

const App = () => {
  return (
    <Printer>
      <SchemaForm
        components={{ Input, Select }}
        onSubmit={values => {
          console.log(values)
        }}
        effects={() => {
          useManyToOneEffects()
        }}
      >
        <Field type="string" name="aa" title="AA" x-component="Input" />
        <Field
          type="string"
          enum={[
            { label: 'visible', value: true },
            { label: 'hidden', value: false }
          ]}
          default={false}
          name="bb"
          title="BB"
          x-component="Select"
        />
        <Field type="string" name="cc" title="CC" x-component="Input" />
      </SchemaForm>
    </Printer>
  )
}

ReactDOM.render(<App />, document.getElementById('root'))
```

**案例解析**

- 多对一联动其实就是一对一联动，只不过作用的对象是同一个字段
- BB 控制 AA 显示隐藏，CC 控制 AA 的值

## 多依赖联动

```jsx
import React, { useEffect } from 'react'
import ReactDOM from 'react-dom'
import {
  SchemaForm,
  SchemaMarkupField as Field,
  FormButtonGroup,
  createFormActions,
  FormEffectHooks,
  createEffectHook,
  Submit,
  Reset
} from '@formily/antd' // 或者 @formily/next
import { combineLatest } from 'rxjs/operators'
import { Input, Select } from '@formily/antd-components'
import Printer from '@formily/printer'
import 'antd/dist/antd.css'

const { onFieldValueChange$, onFormMount$ } = FormEffectHooks

const customEvent$ = createEffectHook('CUSTOM_EVENT')

const useMultiDepsEffects = () => {
  const { setFieldState, dispatch } = createFormActions()

  onFormMount$().subscribe(() => {
    setTimeout(() => {
      dispatch('CUSTOM_EVENT', true)
    }, 3000)
  })

  onFieldValueChange$('aa')
    .pipe(combineLatest(customEvent$()))
    .subscribe(([{ value, values }, visible]) => {
      setFieldState('bb', state => {
        state.visible = visible
      })
      setFieldState('cc', state => {
        state.visible = value
        if (values[1] && values[1].otherInfo) {
          state.value = values[1].otherInfo
        }
      })
    })
}

const App = () => {
  return (
    <Printer>
      <SchemaForm
        components={{ Input, Select }}
        onSubmit={values => {
          console.log(values)
        }}
        effects={() => {
          useMultiDepsEffects()
        }}
      >
        <Field
          type="string"
          enum={[
            { label: 'visible', value: true, otherInfo: '123' },
            { label: 'hidden', value: false, otherInfo: '321' }
          ]}
          default={false}
          name="aa"
          title="AA"
          x-component="Select"
        />
        <Field
          type="string"
          name="bb"
          visible={false}
          title="BB"
          x-component="Input"
        />
        <Field type="string" name="cc" title="CC" x-component="Input" />
      </SchemaForm>
    </Printer>
  )
}

ReactDOM.render(<App />, document.getElementById('root'))
```

**案例解析**

- Field 组件 visible 属性可以控制初始显示状态
- BB 的显示受外部异步事件所控制
- CC 的显示隐藏状态受 AA 的值控制，CC 的值受 AA 的附加信息所控制，同时整体联动依赖一个外部异步事件
- 使用 rxjs 操作符 combineLatest 可以解决联动异步依赖问题

## 链式联动

```jsx
import React, { useEffect } from 'react'
import ReactDOM from 'react-dom'
import {
  SchemaForm,
  SchemaMarkupField as Field,
  FormButtonGroup,
  createFormActions,
  FormEffectHooks,
  Submit,
  Reset
} from '@formily/antd' // 或者 @formily/next
import { Input, Select } from '@formily/antd-components'
import Printer from '@formily/printer'
import 'antd/dist/antd.css'

const { onFieldValueChange$ } = FormEffectHooks

const useChainEffects = () => {
  const { setFieldState } = createFormActions()
  onFieldValueChange$('aa').subscribe(({ value }) => {
    setFieldState('bb', state => {
      state.visible = value
    })
  })
  onFieldValueChange$('bb').subscribe(({ value }) => {
    setFieldState('cc', state => {
      state.visible = value
    })
  })
}

const App = () => {
  return (
    <Printer>
      <SchemaForm
        components={{ Input, Select }}
        onSubmit={values => {
          console.log(values)
        }}
        effects={() => {
          useChainEffects()
        }}
      >
        <Field
          type="string"
          enum={[
            { label: 'visible', value: true },
            { label: 'hidden', value: false }
          ]}
          default={false}
          name="aa"
          title="AA"
          x-component="Select"
        />
        <Field
          type="string"
          enum={[
            { label: 'visible', value: true },
            { label: 'hidden', value: false }
          ]}
          default={false}
          name="bb"
          title="BB"
          x-component="Select"
        />
        <Field type="string" name="cc" title="CC" x-component="Input" />
      </SchemaForm>
    </Printer>
  )
}

ReactDOM.render(<App />, document.getElementById('root'))
```

**案例解析**

- 链式联动，其实也是可以归一化为一对一联动
- AA 控制 BB 显示隐藏，BB 控制 CC 隐藏

## 循环联动

> 联动关系 Total = Price \* Count;Count = Total / Price;Price = Total / Count

```jsx
import React from 'react'
import ReactDOM from 'react-dom'
import {
  SchemaForm,
  SchemaMarkupField as Field,
  FormButtonGroup,
  Submit,
  Reset,
  FormEffectHooks,
  createFormActions
} from '@formily/antd'
import { NumberPicker } from '@formily/antd-components'
import Printer from '@formily/printer'
import 'antd/dist/antd.css'

const { onFormInit$, onFieldValueChange$ } = FormEffectHooks

const useCyclicLinkageEffects = () => {
  const { setFieldState, getFieldState } = createFormActions()
  onFieldValueChange$('total').subscribe(({ value }) => {
    if (!value) return
    setFieldState('count', state => {
      const price = getFieldState('price', state => state.value)
      if (!price) return
      state.value = value / price
    })
    setFieldState('price', state => {
      const count = getFieldState('count', state => state.value)
      if (!count) return
      state.value = value / count
    })
  })
  onFieldValueChange$('price').subscribe(({ value }) => {
    if (!value) return
    setFieldState('total', state => {
      const count = getFieldState('count', state => state.value)
      if (!count) return
      state.value = value * count
    })
    setFieldState('count', state => {
      const total = getFieldState('total', state => state.value)
      if (!total) return
      state.value = total / value
    })
  })
  onFieldValueChange$('count').subscribe(({ value }) => {
    if (!value) return
    setFieldState('total', state => {
      const price = getFieldState('price', state => state.value)
      if (!price) return
      state.value = value * price
    })
    setFieldState('price', state => {
      const total = getFieldState('total', state => state.value)
      if (!total) return
      state.value = total / value
    })
  })
}

const App = () => (
  <Printer>
    <SchemaForm
      components={{
        NumberPicker
      }}
      effects={() => {
        useCyclicLinkageEffects()
      }}
      onChange={v => console.log(v)}
      labelCol={6}
      wrapperCol={4}
      onSubmit={v => console.log(v)}
    >
      <Field
        name="total"
        type="number"
        required
        title="Total"
        x-component="NumberPicker"
      />
      <Field
        name="count"
        type="number"
        required
        title="Count"
        x-component="NumberPicker"
      />
      <Field
        name="price"
        type="number"
        required
        title="Price"
        x-component="NumberPicker"
      />
      <FormButtonGroup offset={6}>
        <Submit />
        <Reset />
      </FormButtonGroup>
    </SchemaForm>
  </Printer>
)
ReactDOM.render(<App />, document.getElementById('root'))
```

**案例详解**

- 循环联动，其实也是可以归一到一对一联动的

## 联动校验

```jsx
import React from 'react'
import ReactDOM from 'react-dom'
import {
  SchemaForm,
  Field,
  FormButtonGroup,
  Submit,
  Reset,
  FormEffectHooks,
  createFormActions
} from '@formily/antd'
import { Input, Password } from '@formily/antd-components'
import Printer from '@formily/printer'
import 'antd/dist/antd.css'

const { onFieldValueChange$ } = FormEffectHooks

const useLinkageValidateEffects = () => {
  const { setFieldState, getFieldState } = createFormActions()
  onFieldValueChange$('*(password,confirm)').subscribe(fieldState => {
    const selfName = fieldState.name
    const selfValue = fieldState.value
    const otherName = selfName == 'password' ? 'confirm' : 'password'
    const otherValue = getFieldState(otherName, state => state.value)
    setFieldState(otherName, state => {
      if (selfValue && otherValue && selfValue !== otherValue) {
        state.errors = '两次密码输入不一致'
      } else {
        state.errors = ''
      }
    })
    setFieldState(selfName, state => {
      if (selfValue && otherValue && selfValue !== otherValue) {
        state.errors = '两次密码输入不一致'
      } else {
        state.errors = ''
      }
    })
  })
}

const App = () => (
  <Printer>
    <SchemaForm
      labelCol={6}
      wrapperCol={6}
      components={{
        Input,
        Password
      }}
      effects={() => {
        useLinkageValidateEffects()
      }}
    >
      <Field
        type="string"
        name="username"
        title="用户名"
        required
        x-component="Input"
      />
      <Field
        type="password"
        name="password"
        title="密码"
        x-props={{ checkStrength: true }}
        description={
          <ul>
            <li>1. 长度不小于8个</li>
            <li>2. 必须包含大小写数字符号</li>
          </ul>
        }
        required
        x-component="Password"
      />
      <Field
        type="password"
        name="confirm"
        title="确认密码"
        x-props={{ checkStrength: true }}
        required
        x-component="Password"
      />
      <FormButtonGroup offset={6}>
        <Submit />
        <Reset />
      </FormButtonGroup>
    </SchemaForm>
  </Printer>
)
ReactDOM.render(<App />, document.getElementById('root'))
```

**案例解析**

- 联动校验都需要手动操作字段状态的 errors 属性来控制，你既需要控制错误的出现时机，也要控制错误的隐藏时机

## 异步联动

```jsx
import React from 'react'
import ReactDOM from 'react-dom'
import {
  SchemaForm,
  SchemaMarkupField as Field,
  FormButtonGroup,
  Submit,
  Reset,
  FormEffectHooks,
  createFormActions,
  FormPath
} from '@formily/antd'
import { Select, Input } from '@formily/antd-components'
import { merge } from 'rxjs'
import Printer from '@formily/printer'
import 'antd/dist/antd.css'

const { onFormInit$, onFieldValueChange$, onFieldInit$ } = FormEffectHooks

const createLinkageUtils = () => {
  const { setFieldState } = createFormActions()
  const linkage = (key, defaultValue) => (path, value) =>
    setFieldState(path, state => {
      FormPath.setIn(state, key, value !== undefined ? value : defaultValue)
    })
  return {
    hide: linkage('visible', false),
    show: linkage('visible', true),
    enum: linkage('props.enum', []),
    loading: linkage('loading', true),
    loaded: linkage('loading', false),
    value: linkage('value')
  }
}

const useAsyncLinkageEffect = () => {
  const linkage = createLinkageUtils()
  onFieldValueChange$('aa').subscribe(fieldState => {
    if (!fieldState.value) return
    linkage.show('bb')
    linkage.loading('bb')
    setTimeout(() => {
      linkage.loaded('bb')
      linkage.enum('bb', ['1111', '2222'])
      linkage.value('bb', '1111')
    }, 1000)
  })
  merge(onFieldValueChange$('bb'), onFieldInit$('bb')).subscribe(fieldState => {
    if (!fieldState.value) return linkage.hide('cc')
    linkage.show('cc')
    linkage.value('cc', fieldState.value)
  })
}

const App = () => (
  <Printer>
    <SchemaForm
      components={{
        Select,
        Input
      }}
      effects={() => {
        useAsyncLinkageEffect()
      }}
      onChange={v => console.log(v)}
      labelCol={6}
      wrapperCol={4}
      onSubmit={v => console.log(v)}
    >
      <Field
        name="aa"
        type="string"
        enum={['aaaaa', 'bbbbb', 'ccccc', 'ddddd', 'eeeee']}
        title="AA"
        x-component="Select"
      />
      <Field
        type="string"
        name="bb"
        title="BB"
        enum={[]}
        visible={false}
        x-component="Select"
        x-props={{
          hasFeedback: true
        }}
      />
      <Field type="string" name="cc" title="CC" x-component="Input" />
      <FormButtonGroup offset={6}>
        <Submit />
        <Reset />
      </FormButtonGroup>
    </SchemaForm>
  </Printer>
)
ReactDOM.render(<App />, document.getElementById('root'))
```

**案例解析**

- 借助 createFormActions，我们可以创建出一些可复用的联动操作原子函数

## 自增列表联动

```jsx
import React from 'react'
import ReactDOM from 'react-dom'
import {
  SchemaForm,
  SchemaMarkupField as Field,
  FormButtonGroup,
  FormEffectHooks,
  createFormActions,
  FormPath,
  Submit,
  Reset
} from '@formily/antd' // 或者 @formily/next
import { ArrayTable, Input, Select } from '@formily/antd-components'
import Printer from '@formily/printer'
import 'antd/dist/antd.css'

const { onFieldValueChange$ } = FormEffectHooks

const App = () => {
  return (
    <Printer>
      <SchemaForm
        components={{ ArrayTable, Input, Select }}
        onSubmit={values => {
          console.log(values)
        }}
        effects={({ setFieldState }) => {
          onFieldValueChange$('array.*.aa').subscribe(({ name, value }) => {
            setFieldState(
              FormPath.transform(name, /\d/, $1 => {
                return `array.${$1}.bb`
              }),
              state => {
                state.visible = value
              }
            )
          })
        }}
      >
        <Field type="array" name="array" x-component="ArrayTable">
          <Field type="object">
            <Field
              type="string"
              name="aa"
              title="Sibling visible"
              enum={[
                { label: '显示', value: true },
                { label: '隐藏', value: false }
              ]}
              default={true}
              x-component="Select"
            />
            <Field type="string" name="bb" title="BB" x-component="Input" />
          </Field>
        </Field>
      </SchemaForm>
    </Printer>
  )
}

ReactDOM.render(<App />, document.getElementById('root'))
```

**案例解析**

- 借助 FormPath.transform 可以求出自增列表字段的相邻字段

## 动态联动

```jsx
import React, { useEffect, useState } from 'react'
import ReactDOM from 'react-dom'
import {
  SchemaForm,
  SchemaMarkupField as Field,
  FormButtonGroup,
  createFormActions,
  FormEffectHooks,
  Submit,
  Reset
} from '@formily/antd' // 或者 @formily/next
import { Input, Select } from '@formily/antd-components'
import Printer from '@formily/printer'
import 'antd/dist/antd.css'

const { onFieldValueChange$ } = FormEffectHooks

const mockSchema = {
  type: 'object',
  properties: {
    aa: {
      type: 'boolean',
      enum: [
        { label: 'visible', value: true },
        { label: 'hidden', value: false }
      ],
      default: false,
      title: 'AA',
      'x-component': 'Select',
      'x-linkages': [
        {
          type: 'value:visible',
          target: '*(bb,cc,dd)',
          condition: '{{!!$value}}'
        }
      ]
    },
    bb: {
      type: 'string',
      title: 'BB',
      'x-component': 'Input'
    },
    cc: {
      type: 'string',
      title: 'CC',
      'x-component': 'Input'
    },
    dd: {
      type: 'string',
      title: 'DD',
      'x-component': 'Input'
    }
  }
}

const App = () => {
  const [schema, setSchema] = useState({
    type: 'object'
  })
  useEffect(() => {
    setTimeout(() => {
      setSchema(mockSchema)
    }, 1000)
  }, [])

  return (
    <Printer>
      <SchemaForm
        schema={schema}
        components={{ Input, Select }}
        onSubmit={values => {
          console.log(values)
        }}
      />
    </Printer>
  )
}

ReactDOM.render(<App />, document.getElementById('root'))
```

- 借助 x-linkages 可以实现简单的联动效果
- target 是一个 FormPathPattern 匹配表达式，在这里我们可以使用 FormPath 的各种匹配语法

## 自增列表动态联动

```jsx
import React, { useEffect, useState } from 'react'
import ReactDOM from 'react-dom'
import {
  SchemaForm,
  SchemaMarkupField as Field,
  FormButtonGroup,
  createFormActions,
  FormEffectHooks,
  Submit,
  Reset
} from '@formily/antd' // 或者 @formily/next
import { Input, Select, ArrayTable } from '@formily/antd-components'
import Printer from '@formily/printer'
import 'antd/dist/antd.css'

const { onFieldValueChange$ } = FormEffectHooks

const mockSchema = {
  type: 'object',
  properties: {
    array: {
      type: 'array',
      'x-component': 'ArrayTable',
      items: {
        type: 'object',
        properties: {
          aa: {
            type: 'string',
            title: 'Sibling visible',
            'x-component': 'Select',
            default: false,
            enum: [
              { label: 'visible', value: true },
              { label: 'hidden', value: false }
            ],
            'x-linkages': [
              {
                type: 'value:state',
                target: '..[].ff',
                state: {
                  visible: '{{!!$value}}'
                }
              }
            ]
          },
          bb: {
            type: 'string',
            title: 'Next Row visible',
            'x-component': 'Select',
            enum: [
              { label: 'visible', value: true },
              { label: 'hidden', value: false }
            ],
            'x-linkages': [
              {
                type: 'value:visible',
                target: 'array.[+].ff',
                condition: '{{!!$value}}'
              }
            ]
          },
          cc: {
            type: 'string',
            title: 'Prev Row visible',
            'x-component': 'Select',
            enum: [
              { label: 'visible', value: true },
              { label: 'hidden', value: false }
            ],
            'x-linkages': [
              {
                type: 'value:visible',
                target: 'array.[-].ff',
                condition: '{{!!$value}}'
              }
            ]
          },
          dd: {
            type: 'string',
            title: 'Next Next Row visible',
            'x-component': 'Select',
            enum: [
              { label: 'visible', value: true },
              { label: 'hidden', value: false }
            ],
            'x-linkages': [
              {
                type: 'value:visible',
                target: 'array.[+2].ff',
                condition: '{{!!$value}}'
              }
            ]
          },
          ee: {
            type: 'string',
            title: 'Prev Prev Row visible',
            'x-component': 'Select',
            enum: [
              { label: 'visible', value: true },
              { label: 'hidden', value: false }
            ],
            'x-linkages': [
              {
                type: 'value:visible',
                target: 'array.[-2].ff',
                condition: '{{!!$value}}'
              }
            ]
          },
          ff: {
            type: 'string',
            title: 'Target',
            'x-component': 'Input'
          }
        }
      }
    }
  }
}

const App = () => {
  const [schema, setSchema] = useState({
    type: 'object'
  })
  useEffect(() => {
    setTimeout(() => {
      setSchema(mockSchema)
    }, 1000)
  }, [])

  return (
    <Printer>
      <SchemaForm
        schema={schema}
        components={{ Input, Select, ArrayTable }}
        onSubmit={values => {
          console.log(values)
        }}
      />
    </Printer>
  )
}

ReactDOM.render(<App />, document.getElementById('root'))
```

**案例解析**

- 自增列表中的动态联动，借助 x-linkages 可以实现相邻联动，或者是跨行联动
- target 相邻查找
  - `prevPath.[].fieldName`代表当前行字段
  - `prevPath.[+].fieldName`代表下一行字段
  - `prevPath.[-].fieldName`代表上一行字段
  - `prevPath.[+2].fieldName`代表下下一行字段
  - `prevPath.[-2].fieldName`代表上上一行字段
  - 一次可以继续往下递增或者递减
- target 向前路径查找
  - `.path.a.b`代表基于当前字段路径往后计算
  - `..path.a.b`代表往前计算相对路径
  - `...path.a.b`代表继续往前计算相对路径
  - 以此类推

## 外部联动

```jsx
import React, { useEffect } from 'react'
import ReactDOM from 'react-dom'
import {
  SchemaForm,
  SchemaMarkupField as Field,
  FormButtonGroup,
  createFormActions,
  FormEffectHooks,
  FormSpy,
  Submit,
  Reset
} from '@formily/antd' // 或者 @formily/next
import { Input, Select } from '@formily/antd-components'
import { Button } from 'antd'
import Printer from '@formily/printer'
import 'antd/dist/antd.css'

const { onFieldValueChange$ } = FormEffectHooks

const useOneToManyEffects = () => {
  const { setFieldState } = createFormActions()
  onFieldValueChange$('aa').subscribe(({ value }) => {
    setFieldState('*(bb,cc,dd)', state => {
      state.visible = value
    })
  })
}

const actions = createFormActions()

const App = () => {
  return (
    <Printer>
      <SchemaForm
        components={{ Input, Select }}
        actions={actions}
        effects={() => {
          useOneToManyEffects()
        }}
      >
        <Field
          type="string"
          enum={[
            { label: 'visible', value: true },
            { label: 'hidden', value: false }
          ]}
          default={false}
          name="aa"
          title="AA"
          x-component="Select"
        />
        <Field type="string" name="bb" title="BB" x-component="Input" />
        <Field type="string" name="cc" title="CC" x-component="Input" />
        <Field type="string" name="dd" title="DD" x-component="Input" />
        <FormButtonGroup>
          <FormSpy selector={[['onFieldValueChange', 'aa']]}>
            {({ state }) => {
              return (
                state.value && (
                  <>
                    <Submit />
                    <Button
                      onClick={() => {
                        actions.setFieldState('bb', state => {
                          state.value = '' + Math.random()
                        })
                      }}
                    >
                      修改BB的值
                    </Button>
                  </>
                )
              )
            }}
          </FormSpy>
        </FormButtonGroup>
      </SchemaForm>
    </Printer>
  )
}

ReactDOM.render(<App />, document.getElementById('root'))
```

**案例解析**

- 主联动逻辑是一对多联动
- 借助 FormSpy 可以针对具体字段做监听，所以可以很方便的做 UI 联动状态同步
- 借助 FormActions 可以方便的在外部操作 Form 内部状态
